import os, sys, sanic, socket

from os import environ as env

from IzzyLib import logging, __version__ as IzzyVersion
from IzzyLib.misc import DefaultDict, DotDict, Path, Boolean, GetIp, RandomGen
from IzzyLib.template import Template
from envbash import load_envbash

from . import __version__ as Version


dbtypes = ['sqlite', 'postgresql', 'mysql', 'mssql']


script_path = Path(__file__).parent()
store_path = script_path.parent(True).join('data')
pyenv = store_path.join('prod.env')


if not store_path.exists():
	store_path.mkdir()


try:
	load_envbash(pyenv.str())

except FileNotFoundError:
	logging.error('Bash environment file not found:', pyenv.str())

	new_config = '''###
# Uncomment and adjust any values as necessary
###

### Port and host PAWS will listen on
#PAWS_HOST=127.0.0.1
#PAWS_PORT=3001

### Set the paws db name at minimum. All other db options will get pulled from mastodon if not set
#PAWS_DB_NAME=paws
#PAWS_DB_HOST=/var/run/postgresql
#PAWS_DB_PORT=5432
#PAWS_DB_USER=mastodon
#PAWS_DB_PASS=
#PAWS_DB_MAXCONNECTIONS=25

### Override domain in mastodon's config
#MASTODOMAIN=mastodon.example.com

### 
#MASTOPATH=/home/mastodon/glitch-soc
#MASTOHOST=localhost:3000
'''

	with pyenv.open('w') as fd:
		fd.write(new_config)

	logging.info('Created a new environment file. Now exiting.')
	sys.exit()


masto_path = Path(env.get('MASTOPATH')) if env.get('MASTOPATH') else None
mastoenv = masto_path.join('.env.production')

if not masto_path:
	logging.error('"MASTOPATH" in environment file is not set')
	sys.exit()

elif not mastoenv.exists():
	logging.error('Cannot find Mastodon\'s ".env.production"')
	sys.exit()

load_envbash(mastoenv.str())


listen = env.get('PAWS_HOST', '127.0.0.1')
port = int(env.get('PAWS_PORT', 3001))
user = env.get('USER', 'paws')

config = DotDict({
	'env': pyenv,
	'path': script_path,
	'data': store_path,
	'frontend': script_path.join('frontend', True),
	'version': Version,
	'dbversion': 20210203,
	'listen': listen,
	'port': port,
	'host': f'{listen}:{port}',
	'salt': env.get('PASS_SALT', RandomGen(40)),
	'secret': env.get('FORWARDED_SECRET', RandomGen(20)),
	'log': DotDict({
		'level': env.get('LOG_LEVEL', 'INFO').upper(),
		'errors': Boolean(env.get('LOG_ERRORS')),
		'date': Boolean(env.get('LOG_DATE', True))
	}),
	'proxy': DotDict({
		'proxy_type': env.get('PROXY_TYPE', 'https'),
		'proxy_host': env.get('PROXY_HOST', None),
		'proxy_port': env.get('PROXY_PORT', 443)
	}),
	'mastodon': DotDict({
		'domain': env.get('LOCAL_DOMAIN'),
		'web_domain': env.get('WEB_DOMAIN', env.get('LOCAL_DOMAIN')),
		'alt_domains': [v for v in env.get('ALTERNATIVE_DOMAINS', '').split(',')],
		'path': masto_path,
		'host': env.get('MASTOHOST', 'localhost:3000')
	}),
	'db': DotDict({
		'host': env.get('PAWS_DB_HOST', env.get('DB_HOST', '/var/run/postgresql')),
		'port': int(env.get('PAWS_DB_PORT', env.get('DB_PORT', 5432))),
		'user': env.get('PAWS_DB_USER', env.get('DB_USER', user)),
		'pass': env.get('PAWS_DB_PASS', env.get('DB_PASS')),
		'name': env.get('PAWS_DB_NAME', 'paws'),
		'maxconnections': int(env.get('PAWS_DB_MAXCONNECTIONS', 25))
	}),
	'mastodb': DotDict({
		'host': env.get('DB_HOST', '/var/run/postgresql'),
		'port': int(env.get('DB_PORT', 5432)),
		'user': env.get('DB_USER', user),
		'password': env.get('DB_PASS'),
		'database': env.get('DB_NAME', 'mastodon_production'),
		'maxconnections': int(env.get('PAWS_DB_MAXCONNECTIONS', 25))
	}),
	'sqfile': store_path.join(env.get('SQ_DATABASE', 'database.sqlite3'))
})

config.agent = f'IzzyLib/{IzzyVersion} (Barkshark-Social/{Version}; +https://{config.mastodon.web_domain}/)',

if not config.log.date:
	log_date = ''

else:
	log_date = '[%(asctime)s] '

logging.setConfig({'level': config.log.level})
config.data.join('template').mkdir()

config.template = Template(
	search=[
		config.data.join('template'),
		config.frontend
	],
	autoescape=False
)

def Context(context, globals):
	config = globals.config
	request = context.request

	try:
		user = request.cookie.user
	except AttributeError:
		user = None

	with globals.mastodb.session() as s:
		data = {
			'cookie_user': user,
			'menu_left': user.config.get('menu_left', False) if user else False,
			'color_theme': request.cookies.get('theme', 'pink'),
			'settings': DefaultDict(),
			'var': {
				'base': 'https://'+request.headers.get('host', config.mastodon.web_domain),
				'name': s.get.config('site_title')
			}
		}

	return data

config.template.func_context = Context
