import sys, traceback

from IzzyLib import logging
from IzzyLib.cache import LRUCache
from IzzyLib.misc import DotDict, RandomGen
from contextlib import contextmanager
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm import sessionmaker

from .config import config
from .database_func import Get, Put, Migrate
from .database_schema import tables, Base
from .database_mastodon import DB
from .functions import Datetime


class DataBase():
	def __init__(self, tables=None, dbtype='postgresql+psycopg2', **kwargs):
		self.engine_string = self.__engine_string(dbtype, kwargs)
		self.db = create_engine(self.engine_string)
		self.table = tables
		self.classes = CustomRows()
		self.session = lambda : Session(self)


	def __engine_string(self, dbtype, kwargs):
		engine_string = dbtype + '://'
		user = kwargs.get('user')
		password = kwargs.get('pass')
		host = kwargs.get('host', '/var/run/postgresql')
		port = kwargs.get('port', 5432)
		name = kwargs.get('name', 'postgres')
		maxconn = kwargs.get('maxconnections', 25)

		if user:
			if password:
				engine_string += f'{user}:{password}@'
			else:
				engine_string += user + '@'

		if host == '/var/run/postgresql':
			engine_string += '/' + name

		else:
			engine_string += f'{host}:{port}/{name}'

		return engine_string


	def CreateDatabase(self):
		if self.engine_string.startswith('postgresql'):
			predb = create_engine(db.engine_string.replace(config.db.name, 'postgres', -1))
			conn = predb.connect()
			conn.execute('commit')

			try:
				conn.execute(f'CREATE DATABASE {config.db.name}')

			except ProgrammingError:
				'The database already exists, so just move along'

			except Exception as e:
				conn.close()
				raise e from None

			conn.close()

		Base.metadata.create_all(self.db)


	def execute(self, *args, **kwargs):
		with self.session() as s:
			return s.execute(*args, **kwargs)


class Session(object):
	def __init__(self, db):
		self.db = db
		self.classes = self.db.classes
		self.session = sessionmaker(bind=db.db)()
		self.table = self.db.table
		self.cache = DotDict()

		# session aliases
		self.s = self.session
		self.commit = self.s.commit
		self.rollback = self.s.rollback
		self.query = self.s.query
		self.execute = self.s.execute

		# extra function aliases
		self.get = Get(self)
		self.put = Put(self)

		for name in self.table.keys():
			self.cache[name] = LRUCache()


	def __enter__(self):
		self.sessionid = RandomGen(10)
		return self


	def __exit__(self, exctype, value, tb):
		if tb:
			self.rollback()
			#print(exctype, value)
			#traceback.print_tb(tb)

		else:
			self.commit()


	def count(self, table_name, **kwargs):
		return self.s.query(self.table[table_name]).filter_by(**kwargs).count()


	def execute(self, string, values=[]):
		return self.s.execute(string, values)


	def fetch(self, table_name, single=True, **kwargs):
		RowClass = self.classes.get(table_name.capitalize())

		rows = self.s.query(self.table[table_name]).filter_by(**kwargs).all()

		if single:
			return RowClass(rows[0], db, self) if len(rows) > 0 else None

		return [RowClass(row, db, self) for row in rows]


	def insert(self, table_name, **kwargs):
		table = self.table[table_name]

		if getattr(table, 'timestamp', None) and not kwargs.get('timestamp'):
			kwargs['timestamp'] = datetime.now()

		data = self.s.add(table(**kwargs))
		return self.fetch(table_name, **kwargs)


	def DropTables(self):
		tables = self.GetTables()

		for table in tables:
			self.s.execute(f'DROP TABLE {table}')


	def GetTables(self):
		rows = self.s.execute("SELECT name FROM sqlite_master WHERE type IN ('table','view') and name NOT LIKE 'sqlite_%'")
		return [row[0] for row in rows]


class CustomRows(object):
	def get(self, name):
		return getattr(self, name, self.Row)


	class Row(DotDict):
		_filter_columns = lambda self, row: [attr for attr in dir(row) if not attr.startswith('_') and attr != 'metadata']


		def __init__(self, row, db, s):
			if not row:
				return

			self._db = db
			self._table_name = row.__tablename__
			self._table = db.table[row.__tablename__]
			self._columns = self._filter_columns(row)

			super().__init__()

			for k in self._columns:
				v = getattr(row, k)

				if type(v) == datetime:
					v = Datetime(v)

				if type(v) == dict:
					v = DotDict(v)

				self[k] = v

			self.__run__(s)


		## Subclass Row and redefine this function
		def __run__(self, s):
			pass


		def _filter_data(self):
			data = {k: v for k,v in self.items() if k in self._columns}

			for k,v in self.items():
				if v.__class__ == DotDict:
					data[k] = v.asDict()

			return data


		def asDict(self):
			return self._filter_data()


		def _update(self, new_data={}, **kwargs):
			kwargs.update(new_data)

			for k,v in kwargs.items():
				self[k] = v


		def delete(self):
			with self._db.session() as s:
				row = s.query(self._table).filter_by(id=self.id)
				row.delete()

				return True if not row.all() else False


		def delete_session(self, s):
			row = s.query(self._table).filter_by(id=self.id)
			row.delete()

			return True if not row.all() else False


		def update(self, dict_data={}, **data):
			self._update(data)
			self._update(dict(dict_data))

			data = self._filter_data()

			with self._db.session() as s:
				s.query(self._table).filter_by(id=self.id).update(data)


		def update_session(self, s, dict_data={}, **data):
			self._update(data)
			self._update(dict_data)

			data = self._filter_data()

			row = s.query(self._table).filter_by(id=self.id)
			row.update(data)


db = DataBase(tables, **config.db.asDict())
mastodb = DB(**config.mastodb.asDict())

try:
	with db.session() as s:
		config_version = s.get.config('version', 0)
except (OperationalError, ProgrammingError):
	config_version = 0

if config_version < config.dbversion:
	Migrate(config_version, db)
