import sys, traceback

from IzzyLib import logging
from IzzyLib.misc import Boolean, DefaultDict, DotDict, RandomGen
from datetime import datetime, timedelta
from sqlalchemy_paginator import Paginator

from .config import config
from .functions import Datetime


subtypes = {
	'str': str,
	'int': int,
	'bool': Boolean,
	'datetime': int
}


config_defaults = {
	'version': [config.dbversion, 'int'],
	'require_allow': [False, 'bool']
}


def SetupClass(self, session):
	self.session = session
	self.s = session.s
	self.classes = session.db.classes
	self.db = session.db

	self.cache = session.cache
	self.count = session.count
	self.fetch = session.fetch
	self.insert = session.insert
	self.table = session.table

	self.query = session.s.query
	self.execute = session.s.execute


class Get():
	def __init__(self, session):
		SetupClass(self, session)


	def __parse_config(self, row):
		typefunc = subtypes[row.type] if row.type else str
		value = typefunc(row.value)

		if row.key == 'emoji_size_limit':
			value = value * 1024

		return Datetime(datetime.fromtimestamp(value)) if row.type == 'datetime' else value


	def config(self, key, default=None, cache=True):
		cached = self.cache.config.fetch(key)

		if cached and cache:
			return cached

		row = self.fetch('config', key=key)

		if row:
			var = self.__parse_config(row)
			self.cache.config.store(key, var)

			return var

		if not default:
			value = config_defaults.get(key)
			return value * 1000 if key == 'emoji_size_limit' else value

		return default


	def configs(self, cache=True):
		data = DotDict()

		for key, value in config_defaults.items():
			data[key] = self.config(key, value, cache=cache)

		return data


	def cookie(self, cache=True):
		return None


class Put():
	def __init__(self, session):
		SetupClass(self, session)


	def config(self, key, value, subtype='str'):
		row = self.fetch('config', key=key)
		typefunc = subtypes[row.type] if row else str

		if subtype == 'bool':
			new_value = 'True' if Boolean(value) else 'False'

		else:
			new_value = int(value.timestamp()) if subtype == 'datetime' else typefunc(value)

		if row:
			if row.value == new_value:
				return

			row.value = new_value
			row.update_session(self.s)

		else:
			self.insert('config',
				key=key,
				value=new_value,
				type=subtype
			)

		self.cache.config.remove(key)


	def configs(self, *args):
		for config in args:
			key = config.get('key')
			value = config.get('value')
			subtype = config.get('subtype')

			if None in [key, value]:
				logging.warning(f'db.put.configs: Missing key or value for config: key={key}, value={value}')
				continue

			self.config(key, value, subtype)


def Setup(db):
	try:
		db.CreateDatabase()
	except Exception as e:
		traceback.print_exc()
		logging.error('Failed to create database. Exiting...')
		sys.exit()

	with db.session() as s:
		for k,v in config_defaults.items():
			s.put.config(k, v[0], v[1])


def Migrate(dbversion, db):
	if dbversion == 0:
		Setup(db)
		return

	logging.info('Migrated database to version', config.dbversion)
