import psycopg2, traceback, sys, json

from IzzyLib import logging
from IzzyLib.misc import Boolean, DotDict, Path
from contextlib import contextmanager
from datetime import datetime
from dbutils.pooled_db import PooledDB


class DB(PooledDB):
	def __init__(self, **kwargs):
		kwargs['maxusage'] = 1
		super().__init__(psycopg2, **kwargs)

		self.session = lambda : Session(self)


	def query(self, string, args: list=[]):
		with self.session() as s:
			return self.query(string, args)


	def fetch(self, *args, **kwargs):
		with self.session(False) as s:
			return self.fetch(*args, **kwargs)


class Session(object):
	def __init__(self, db, trans=False):
		self.db = db
		self.trans = trans
		self.conn = None
		self.cursor = None
		self.get = Get(self)
		self.put = Put(self)


	def __enter__(self):
		self.conn = self.db.connection()

		if self.trans:
			self.conn.begin()
		self.cursor = self.conn.cursor()

		return self


	def __exit__(self, exctype, value, tb):
		self.cursor.close()

		if self.trans:
			if tb:
				self.conn.rollback()
				#print(exctype, value)
				#traceback.print_tb(tb)

			else:
				self.commit()

		self.close()


	def close(self):
		self.conn.close()


	def query(self, string, values=[]):
		values = list(values)
		self.cursor.execute(string, values)
		return self.cursor


	def fetch(self, table, single=True, sort=None, **kwargs):
		rowid = kwargs.get('id')
		querysort = f'ORDER BY {sort}'

		resultOpts = [self, table, self.cursor]

		if rowid:
			rows = self.query(f"SELECT * FROM {table} WHERE id = ?", [rowid])

		elif kwargs:
			placeholders = [f'{k} = %s' for k in kwargs.keys()]
			values = kwargs.values()

			where = ' and '.join(placeholders)
			query = f"SELECT * FROM {table} WHERE {where} {querysort if sort else ''}"
			rows = self.query(query, values)

		else:
			rows = self.query(f'SELECT * FROM {table} {querysort if sort else ""}')

		rows = rows.fetchall() if not single else rows.fetchone()

		if rows:
			if single:
				return DBResult(rows, *resultOpts)

			return [DBResult(row, *resultOpts) for row in rows]

		return None if single else []


	def insert(self, table, data={}, **kwargs):
		data.update(kwargs)
		placeholders = ",".join(['?' for _ in data.keys()])
		values = tuple(data.values())
		keys = ','.join(data.keys())

		#if 'timestamp' in self.main.tables[table].keys() and 'timestamp' not in keys:
			#data['timestamp'] = datetime.now()

		query = f'INSERT INTO {table} ({keys}) VALUES ({placeholders})'
		self.query(query, values)
		return True


	def remove(self, table, **kwargs):
		keys = []
		values = []

		for k,v in kwargs.items():
			keys.append(k)
			values.append(v)

		keydata = ','.join([f'{k} = ?' for k in keys])
		query = f'DELETE FROM {table} WHERE {keydata}'

		self.query(query, values)


	def update(self, table, rowid, data={}, **kwargs):
		data.update(kwargs)
		newdata = {k: v for k, v in data.items() if k in self.main.tables[table].keys()}
		keys = list(newdata.keys())
		values = list(newdata.values())

		if len(newdata) < 1:
			logging.debug('No data provided to update row')
			return False

		query_data = ', '.join(f'{k} = ?' for k in keys)
		query = f'UPDATE {table} SET {query_data} WHERE id = {rowid}'

		self.query(query, values)


	def count(self, table):
		data = self.query(f'SELECT COUNT(*) FROM {table}')
		return data.fetchone()[0]


class Get():
	def __init__(self, session):
		ParseSession(self, session)


	def config(self, key, default=None):
		row = self.fetch('settings', var=key)

		if not row:
			return default

		return row.value[4:]


	def configs(self):
		data = DotDict()

		for row in self.fetch('settings', single=False):
			data[row.var] = row.value

		return data


class Put():
	def __init__(self, session):
		ParseSession(self, session)


class DBResult(DotDict):
	def __init__(self, row, db, table, cursor):
		super().__init__()

		self.db = db
		self.table = table
		self.columns = []

		for idx, col in enumerate(cursor.description):
			self.columns.append(col[0])
			self[col[0]] = row[idx]


	def __str__(self):
		return self.json()


	def _filter_data(self):
		data = {k: v for k,v in self.items() if k in self.columns}

		for k,v in data.items():
			if v.__class__ == DotDict:
				data[k] = v.asDict()

		return data


	def Dict(self):
		return self._filter_data()


	def Json(self, indent=None):
		data = self.asDict()

		for k,v in data.items():
			if v.__class__ == datetime:
				data[k] = str(v)

		return json.dumps(data, indent=indent)


	def Update(self, data={}):
		with self.db.session(True) as s:
			self.update(data)
			s.update(self.table, self.id, self.AsDict())


	def Remove(self):
		with self.db.session(True) as s:
			s.remove(self.table, id=self.id)


def ParseSession(self, session):
	self.session = session
	self.db = session.db

	# Basic function aliases
	self.query = session.query
	self.fetch = session.fetch
	self.insert = session.insert
	self.remove = session.remove
	self.update = session.update
	self.count = session.count
