import functools, json, os, pytz, re, sys, socket

from IzzyLib import logging, color
from IzzyLib.http import HttpClient, SetClient
from IzzyLib.misc import DefaultDict, DotDict
from base64 import b64decode, b64encode
from sanic import response
from datetime import datetime
from hashlib import md5
from pytz import timezone
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from .config import config


themes = {
	'blue': {'primary': '#55e'},
	'green': {'primary': '#2e5'},
	'orange': {'primary': '#e52'},
	'red': {'primary': '#e24'},
	'pink': {'primary': '#e29'},
	'purple': {'primary': '#a2e'},
	'yellow': {'primary': '#ed2'}
}


def mkhash(string, salt=config.salt, alg='blake'):
	hashalg = hashalgs.get(alg.lower())

	if not hashalg:
		raise KeyError(f'Not a valid hash algorithm: {alg}')

	string = string.encode('UTF-8') if type(string) != bytes else string
	salt = salt.encode('UTF-8') if type(salt) != bytes else salt

	newhash = hashalg() if alg == 'md5' else hashalg.new()
	newhash.update(string)
	newhash.update(salt)
	return newhash.hexdigest()




def GetTheme(name):
	defaults = {
		'primary': '#e7a',
		'secondary': '#e7a',
		'background': '#191919',
		'positive': '#ada',
		'negative': '#daa'
	}

	defaults.update(themes.get(name, {}))
	return defaults


def CssTimestamp(*args):
	timestamp = 0

	for name in args:
		filename = config.frontend.join(f'{name}.css')
		timestamp += int(filename.mtime())

	return timestamp


def Error(request, code, msg=None):
	message = msg if msg else ''

	if any(map(request.path.startswith, ['/status', '/actor'])) or JsonCheck(request):
		return JsonResp({'error': message}, code)

	else:
		cont_type = 'text/html'
		data = {
			'login_token': request.cookies.get('login_token') if not isinstance(request, str) else '',
			'code': str(code),
			'msg': msg
		}

		return config.template.response(f'error.haml', request, {'data': data, 'msg': message}, status=code)


def FetchActor(url, user=None):
	url = url.split('#')[0]
	headers = {'Accept': 'application/activity+json'}

	if user:
		resp = Client.signed_request(user.privkey, f'https://{config.web_domain}/actor/{user.handle}', url, headers=headers)

	else:
		resp = Client.request(url, headers=headers)

	if not resp.json():
		logging.verbose('functions.FetchActor: Failed to fetch actor:', url)
		logging.debug(f'Error {resp.status}: {resp.body}')
		return

	actor = resp.json()
	actor.web_domain = urlparse(url).netloc
	actor.shared_inbox = actor.inbox
	actor.pubkey = None
	actor.handle = actor.preferredUsername

	if actor.get('endpoints'):
		actor.shared_inbox = actor.endpoints.get('sharedInbox', actor.inbox)

	if actor.get('publicKey'):
		actor.pubkey = actor.publicKey.get('publicKeyPem')

	return actor


def FetchApi(domain):
	return Client.request(f'https://{domain}/api/v1/instance').json()


@functools.lru_cache(maxsize=2048)
def FetchWebfingerAcct(handle, domain):
	data = DefaultDict({'actor': None})
	webfinger = Client.request(f'https://{domain}/.well-known/webfinger?resource=acct:{handle}@{domain}')

	if not webfinger.body:
		return

	data.handle, data.domain = webfinger.json().subject.replace('acct:', '').split('@')

	for link in webfinger.json().links:
		if link['rel'] == 'self' and link['type'] == 'application/activity+json':
			data.actor = link['href']

	return data


def JsonCheck(request):
	if any(map(request.path.startswith, ['/actor', '/inbox', '/status'])) or request.path.endswith('.json'):
		return True

	accept = request.headers.get('Accept')

	if not accept:
		return

	mimes = accept.split(',')

	if any(mime in ['application/json', 'application/activity+json'] for mime in mimes):
		return True

	return False


def JsonResponse(data, status=200, headers=None, activity=False):
	kwargs = {
		'status': status
	}

	if activity:
		kwargs['content_type'] = 'application/activity+json'

	if headers:
		kwargs['headers'] = headers

	return response.json(json.dumps(data), **kwargs)


class Datetime(object):
	def __init__(self, date: datetime=None):
		self.date = date if date else datetime.utcnow()


	def __str__(self):
		return f'{self.string("%Y-%m-%d %H:%M:%S")}'


	def ap_string(self):
		return self.string('%Y-%m-%dT%H:%M:%SZ')


	def string(self, format):
		return self.date.strftime(format)


	def timestamp(self, integer=True):
		ts = self.date.timestamp()
		return int(ts) if integer else ts


	def tz(self, timezone):
		return pytz.timezone(timezone).localize(self.date)


Client = HttpClient(useragent=config.agent, **config.proxy.asDict(), timeout=5)
SetClient(Client)
