import binascii, base64, json, hashlib, multiprocessing, traceback

from urllib.parse import quote, urlparse, unquote

from IzzyLib import logging
from IzzyLib.cache import LRUCache
from IzzyLib.http import ParseSig, PkcsHeaders, VerifyRequest
from IzzyLib.misc import DefaultDict, DotDict
from base64 import b64encode, b64decode
from sanic.request import Request as SanicRequest
from sanic.response import raw, redirect, json as rjson
from sanic import exceptions as exc
from tldextract import extract
from urllib.parse import parse_qsl

from .config import config
from .database import db
from .errors import Teapot
from .functions import Client, FetchActor, FetchWebfingerAcct


blocked_agents = [
	'gabsocial',
	'soapbox',
	'spinster',
	'kiwifarms'
]

domain_bans = [
	'b77625108b127dc1b15640ac536c8e2991b83ea20841a42e70927f96434fe9c2',
	'f1e9527ea188da343fe79f23855f80476cedb017e3880432ab22467eaf992f44',
	'a67a3f85d90881ae32716d0eaf5a348b53096e07ea38a20797d74824074e281b',
	'75e80dd194bc635186cf0b4a07524d7746b4b261cbd5fb15007189e4fccfbbb6',
	'56638777979b0e88cd6cd67723f4fc152eae7ada5c266cdab45e8bb29af6be24',
	'acb2a6826ab0741142935b0e0454b381c0b32a5dfcf54bd43c354379263aec1f',
	'e249dc1503a399ad9fb807ded9be6dd42ca9ea04fa04adb8f794435e1ddb8554',
	'feea7dc93ff4b6b7c06bc3d2dbd6868b259dccd664309c859ed56d3480971617',
	'f0f7b4b28f4142e276d39b1fddfe621e81f72d3be86296a6323d303164f423a2',
	'bfaa1ed5a206867a79f7fbd187f98e035a1cee5fa1e3821740dfc0a9cb3d6ecc',
	'0dbb3144d7b6dff976d97a54be2e94da3559d089eb5c00f871e18318afb4f047',
	'56da8b2c7030ff79b9acba7a23fb6750a165a88505e0dcac85478d537d88e90f',
	'6dd58fd367069c428c481182a8a6f62af21b7a33207b12cb15140da06b8401af',
	'e7c641cf9cd8b5c4513e22d0dfd3f914835a16bd4011801dfa0976fedf61bf02',
	'a3da30d2dfb76009289509cb3f35abd0a43b0736f0abf1ddd17c05685902a6fb',
	'19452ff02670099643197b3fec15c0d43e42a3dbbe83b0fb4d039d5f77d1932e',
	'e8bd7afb2125b7ea5e1b885fac2a657d817eec289ed870aa3d1d0acbe6aded8f'
]

blocked_domains = [extract(domain) for domain in domain_bans]
hash_cache = LRUCache(4096)

paths = DotDict({
	'anon_api': [
		'/api/v1/apps',
		'/api/v1/instance'
	],
	'ap': [
		'/status',
		'/actor',
		'/inbox'
	],
	'cookie': [
		'/@',
		'/:',
		'/welcome',
		'/preferences',
		'/admin',
		'/oauth/authorize'
	],
	'log_ignore': [
		'/media',
		'/static',
		'/style',
		'/manifest',
		'/favicon'
	]
})


async def http_bans(request):
	# Let the request go through if it's basic resources needed for the error page
	if request.paths.media:
		return

	if [agent for agent in blocked_agents if agent in request.headers.get('User-Agent', '').lower()]:
		raise Teapot('This teapot kills fascists')

	domain_hash = hash_cache.fetch(request.ap.top_domain)

	if not domain_hash:
		domain = request.ap.top_domain
		domain_hash = hashlib.sha256(domain.encode("utf-8")).hexdigest()
		hash_cache.store(domain, domain_hash)

	if domain_hash in blocked_domains:
		raise Teapot('This teapot kills fascists')


async def http_auth(request):
	if request.paths.media:
		return

	if request.ap.web_domain == config.web_domain:
		return rjson({'error': 'unwilling to handle self'}, status=403)

	if not request.paths.api_anon and request.paths.api:
		resp = handle_api_path(request)

		if resp:
			return resp

	if request.paths.ap:
		resp = handle_ap_path(request)

		if resp:
			return
			return resp

	if request.paths.cookie:
		resp = handle_web_path(request)

		if resp:
			return resp


async def http_headers(request, response):
	if not response.headers.get('Cache-Control'):
		if request.path.startswith('/system') or any(map(request.path.endswith, ['woff2'])):
			response.headers['Cache-Control'] = 'public,max-age=2628000,immutable'

		else:
			response.headers['Cache-Control'] = 'no-store'

	response.headers['Server'] = f'PAWS/{config.version}'
	response.headers['Trans'] = 'Rights'


async def http_access_log(request, response):
	if request.paths.log_ignore or response.status == 302:
		return

	uagent = request.headers.get('user-agent')
	address = request.forwarded.get('for', request.remote_addr)

	logging.info(f'({multiprocessing.current_process().name}) {address} {request.method} {request.path} {response.status} "{uagent}"')


def handle_web_path(request):
	with db.session() as s:
		handle = request.match_info.get('user')
		statusid = request.match_info.get('statusid')

		if any(map(request.path.startswith, ['/@', '/:'])) and request.paths.json:
			return

		if request.path.startswith('/@'):
			domainid = s.get.domainid(config.domain)
			user = s.fetch('user', handle=handle, domainid=domainid)

			if not user or user.config.get('public'):
				return

	if not request.cookie.token:
		if request.paths.json:
			return rjson({'error': 'Missing token'}, status=401)

		else:
			return redirect(f'/login?redir={quote(request.path)}&query={quote(request.query_string)}')

	elif not request.cookie.token:
		return redirect(f'/login?redir={quote(request.path)}&msg=InvalidToken')


def handle_api_path(request):
	if not request.api.token or not request.api.user:
		logging.debug('middleware.http_auth: Invalid token:', request.api.token)
		return rjson({'error': 'Invalid token'}, status=401)


def handle_ap_path(request):
	if not request.verify():
		logging.debug('middleware.http_auth: Signature verify failed')

		if request.method == 'GET' and request.path.startswith('/actor') and len(request.path.split('/')) == 3:
			pass

		else:
			return rjson({'error': 'Failed to verify signature'}, status=401)

class InvalidAttr(object):
	pass
