import json

from IzzyLib import logging
from IzzyLib.cache import LRUCache
from IzzyLib.http import ParseSig, PkcsHeaders, VerifyHeaders, VerifyRequest
from IzzyLib.misc import DotDict
from sanic.request import Request
from sanic.response import raw, redirect, json as rjson
from tldextract import extract
from urllib.parse import parse_qsl, quote, urlparse, unquote

from .config import config
from .database import db
from .errors import Teapot
from .functions import Client, FetchActor, FetchWebfingerAcct


paths = DotDict({
	'anon_api': [
		'/api/v1/apps',
		'/api/v1/instance'
	],
	'ap': [
		'/status',
		'/actor',
		'/inbox'
	],
	'cookie': [
		'/@',
		'/:',
		'/welcome',
		'/preferences',
		'/admin',
		'/oauth/authorize'
	],
	'log_ignore': [
		'/media',
		'/static',
		'/style',
		'/manifest',
		'/favicon'
	],
	'log_ext_ignore': [
		'png',
		'svg',
		'js',
		'jpg',
		'css',
		'woff2',
		'ttf'
	]
})


class Headers(DotDict):
	def __init__(self, headers: dict):
		headers_lower = {k.lower(): v for k,v in headers.items()}
		super().__init__(headers_lower)


	def __setitem__(self, k, v):
		DotDict.__setitem__(self, k.lower(), v)


	def __repr__(self):
		return ', '.join([f'{k}={v}' for k,v in self.items()])


	def getall(self, k, default=None):
		return self.get(k.lower(), default)


class Data(object):
	def __init__(self, request):
		self.request = request


	@property
	def combined(self):
		data = DotDict()

		for item in [self.form, self.query, self.json]:
			data.update(item)

		return data


	@property
	def query(self):
		data = {k: v for k,v in parse_qsl(self.request.query_string)}
		return DotDict(data)


	@property
	def form(self):
		data = {k: v[0] for k,v in self.request.form.items()}
		return DotDict(data)


	@property
	def files(self):
		return DotDict({k:v[0] for k,v in self.request.files.items()})


	### body functions
	@property
	def raw(self):
		try:
			return self.request.body
		except Exception as e:
			logging.error('middleware.Content.raw: failed to get body')
			logging.debug(f'{e.__class__.__name__}: {e}')
			return b''


	@property
	def text(self):
		try:
			return self.raw.decode()
		except Exception as e:
			logging.error('middleware.Content.text: failed to get body')
			logging.debug(f'{e.__class__.__name__}: {e}')
			return ''


	@property
	def json(self):
		try:
			return DotDict(self.text)
		except Exception as e:
			logging.error('middleware.Content.json: failed to get body')
			logging.debug(f'{e.__class__.__name__}: {e}')
			data = '{}'
		return {}


class HTTPRequest(Request):
	def __init__(self, url_bytes, headers, version, method, transport, app):
		super().__init__(url_bytes, headers, version, method, transport, app)

		self.headers = Headers(headers)
		self.data = Data(self)
		self.cookie = DotDict({'token': None, 'user': None})
		self.__parse_path()

		if self.paths.media:
			return

		with db.session() as s:
			self.__parse_cookie(s)


	def alldata(self):
		return self.__combine_dicts(self.data.json, self.data.query, self.data.form)


	def __try_func(self, func, *args, default=None, **kwargs):
		try:
			return func(*args, *kwargs)
		except Exception as e:
			return default


	def __combine_dicts(self, *dicts):
		data = DotDict()

		for item in dicts:
			data.update(item)

		return data


	def __parse_path(self):
		self.paths = DotDict()
		self.paths.media = self.path.startswith('/paws/static')
		self.paths.log_ignore = any(map(self.path.endswith, paths.log_ext_ignore))
		self.paths.api = self.path.startswith('/api')
		self.paths.api_anon = any(map(self.path.startswith, paths.anon_api))
		self.paths.ap = any(map(self.path.startswith, paths.ap))
		self.paths.cookie = any(map(self.path.startswith, paths.cookie))
		self.paths.json = self.__json_check()


	def __json_check(self):
		if any(map(self.path.startswith, paths.ap)) or self.path.endswith('.json'):
			return True

		accept = self.headers.get('Accept')

		if accept:
			mimes = accept.split(',')

			if any(mime in ['application/json', 'application/activity+json'] for mime in mimes):
				return True

		return False


	def __parse_cookie(self, s):
		login_token = self.cookies.get('login_token')
		cookie = s.get.cookie(login_token)

		if login_token and cookie:
			self.cookie = cookie
