import asyncio, signal, sys, json, os

import sanic

from IzzyLib import logging
from IzzyLib.misc import GetIp
from sanic import exceptions as exc
from jinja2.exceptions import TemplateNotFound
from sqlalchemy.orm.exc import DetachedInstanceError
from uuid import uuid4

from . import views, errors, middleware
from .database import db, mastodb
from .config import config
from .functions import CssTimestamp
from .errors import Teapot
from .request import HTTPRequest


config.template.updateEnv({
	'config': config,
	'db': db,
	'mastodb': mastodb,
	'CssTimestamp': CssTimestamp
})

web = sanic.Sanic('social', request_class=HTTPRequest)
web.config['FORWARDED_SECRET'] = config.secret


def safe_stop(*args):
	#logging.debug('Saving updated timestamps')
	#update_timestamps()

	logging.info('Bye')
	sys.exit()


def setupRoutes():
	error_handlers = [
		(exc.NotFound, errors.generic),
		(exc.MethodNotSupported, errors.generic),
		(exc.InvalidUsage, errors.generic),
		(exc.Unauthorized, errors.generic),
		(exc.Forbidden, errors.generic),
		(Teapot, errors.generic),
		(TemplateNotFound, errors.missing_template),
		(DetachedInstanceError, errors.orm_detached_instance),
		(Exception, errors.server_error)
	]

	for handler in error_handlers:
		web.error_handler.add(*handler)

	middlewares = [
		#(middleware.http_domain_user_check, 'request'),
		#(middleware.http_bans, 'request'),
		#(middleware.http_auth, 'request'),
		(middleware.http_headers, 'response'),
		(middleware.http_access_log, 'response')
	]

	for m, a in middlewares:
		web.register_middleware(m, attach_to=a)

	web_routes = [
		(views.MastoBase, '/'),
		(views.MastoBase, '/<path:[A-z0-9_\-/@:%.()&$#@]+>'),
		(views.MastoUser, '/@<user>'),
		(views.MastoStatus, '/:<statusid:int>')
	]

	for view, route in web_routes:
		web.add_route(view, route, methods=['GET', 'POST', 'HEAD', 'PUT'])


	paws_routes = [
		(views.Home, '/paws'),
		(views.Settings, '/paws/settings'),
		(views.Instances, '/paws/instances'),
		(views.Login, '/paws/login'),
		(views.Logout, '/paws/logout'),
		(views.Style, '/paws/style-<timestamp:int>.css'),
	]

	for view, route in paws_routes:
		web.add_route(view.as_view(), route)


	# Media
	web.static('/paws/static', config.frontend.join('static').str())
	web.static('/packs', config.mastodon.path.join('public/packs').str())


def start():
	setupRoutes()

	signal.signal(signal.SIGHUP, safe_stop)
	signal.signal(signal.SIGINT, safe_stop)
	signal.signal(signal.SIGQUIT, safe_stop)
	signal.signal(signal.SIGTERM, safe_stop)

	listen = config.listen
	port = config.port
	workers = os.cpu_count()

	logging.info(f'Starting PAWS at {listen}:{port}')
	try:
		web.run(
			host=listen,
			port=port,
			workers=workers if workers else 1,
			debug=False,
			access_log=False
		)

	except Exception as e:
		print('server exit error:', e.__class__.__name__, e)

	safe_stop()
